module UU
  module OS
    module Lang

      # Implementation of {Hash} providing methods for unification of stored keys and values
      # on both, persistence level (all values are stored in same form) and API level (any
      # user input is transformed to match internal representation). By default, all keys
      # are translated to symbols, values are left without any change.
      class UniformHash < ::Hash

        alias_method :_has_key?, :has_key?
        alias_method :_put, :[]=
        alias_method :_get, :[]
        private :_has_key?, :_put, :_get

        # @see Hash#initialize
        def initialize(obj = nil)
          super(obj)
        end

        # @see Hash#[]
        def [](key)
          super(transform_key(key))
        end

        # @see Hash#[]=
        def []=(key, value)
          super(transform_key(key), transform_value(value))
        end

        # @see Hash#default
        def default(key = nil)
          super(transform_key(key))
        end

        # @see Hash#delete
        def delete(key, &block)
          super(transform_key(key), &block)
        end

        # @see Hash#fetch
        def fetch(key, default = nil, &block)
          super(transform_key(key), default, &block)
        end

        # @see Hash#has_key?
        def has_key?(key)
          super(transform_key(key))
        end

        # @see Hash#has_value?
        def has_value?(value)
          super(transform_value(value))
        end

        # @see Hash#include?
        def include?(key)
          super(transform_key(key))
        end

        # @see Hash#key?
        def key?(key)
          super(transform_key(key))
        end

        # @see Hash#key
        def key(value)
          super(transform_value(value))
        end

        # @see Hash#member?
        def member?(key)
          super(transform_key(key))
        end

        # @see Hash#merge
        def merge(other_hash, &block)
          result = self.class.new
          result.replace(self)
          result.merge!(other_hash, &block) if other_hash
          return result
        end

        # @see Hash#merge!
        def merge!(other_hash, &block)
          if other_hash
            raise TypeError.new("no implicit conversion of #{other_hash.class} into Hash") if !other_hash.respond_to?:each
            other_hash.each do |key, value|
              tkey = transform_key(key)
              if !_has_key?(tkey) || block.nil?
                _put(tkey, transform_value(value))
              else
                _put(tkey, transform_value(block.yield(tkey, _get(tkey), value)))
              end
            end
          end
        end

        # @see Hash#replace
        def replace(other_hash)
          if other_hash
            raise TypeError.new("no implicit conversion of #{other_hash.class} into Hash") if !other_hash.respond_to?:each
            clear
            other_hash.each do |key, value|
              self[key] = value
            end
          else
            clear
          end
        end

        # @see Hash#store
        def store(key, value)
          super(transform_key(key), transform_value(value))
        end

        # @see Hash#update
        def update(other_hash, &block)
          merge!(other_hash, &block) if other_hash
        end

        # @see Hash#value?
        def value?(value)
          super(transform_value(value))
        end

        # @see Hash#values_at
        def values_at(*key)
          if (key.size == 0)
            super
          else
            keys = []
            key.each do |k|
              keys << transform_key(k)
            end
            super(keys)
          end
        end

        # Transforms given key to desired form.
        # @param [Object] key Key to be transformed.
        # @return [Object] Transformed key.
        # @note Not intended to be directly used, but for overriding
        #    of default behavior by custom implementation!
        def transform_key(key)
          return nil if key.nil?
          return key.to_sym
        end

        # Transforms given value to desired form.
        # @param [Object] value Value to be transformed.
        # @return [Object] Transformed value.
        # @note Not intended to be directly used, but for overriding
        #    of default behavior by custom implementation!
        def transform_value(value)
          return value
        end

      end

    end
  end
end
